from typing import Optional

from frostfs_testlib.cli.cli_command import CliCommand
from frostfs_testlib.shell import CommandResult


class FrostfsCliControl(CliCommand):
    def set_status(
        self,
        endpoint: str,
        status: str,
        wallet: Optional[str] = None,
        force: Optional[bool] = None,
        address: Optional[str] = None,
        timeout: Optional[str] = None,
    ) -> CommandResult:
        """Set status of the storage node in FrostFS network map

        Args:
        wallet: Path to the wallet or binary key
        address: Address of wallet account
        endpoint: Remote node control address (as 'multiaddr' or '<host>:<port>')
        force: Force turning to local maintenance
        status: New netmap status keyword ('online', 'offline', 'maintenance')
        timeout: Timeout for an operation (default 15s)

        Returns:
            Command`s result.
        """
        return self._execute(
            "control set-status",
            **{param: value for param, value in locals().items() if param not in ["self"]},
        )

    def healthcheck(
        self,
        endpoint: str,
        wallet: Optional[str] = None,
        address: Optional[str] = None,
        timeout: Optional[str] = None,
    ) -> CommandResult:
        """Health check for FrostFS storage nodes

        Args:
        wallet: Path to the wallet or binary key
        address: Address of wallet account
        endpoint: Remote node control address (as 'multiaddr' or '<host>:<port>')
        timeout: Timeout for an operation (default 15s)

        Returns:
            Command`s result.
        """
        return self._execute(
            "control healthcheck",
            **{param: value for param, value in locals().items() if param not in ["self"]},
        )

    def drop_objects(
        self,
        endpoint: str,
        objects: str,
        wallet: Optional[str] = None,
        address: Optional[str] = None,
        timeout: Optional[str] = None,
    ) -> CommandResult:
        """Drop objects from the node's local storage

        Args:
        wallet: Path to the wallet or binary key
        address: Address of wallet account
        endpoint: Remote node control address (as 'multiaddr' or '<host>:<port>')
        objects: List of object addresses to be removed in string format
        timeout: Timeout for an operation (default 15s)

        Returns:
            Command`s result.
        """
        return self._execute(
            "control drop-objects",
            **{param: value for param, value in locals().items() if param not in ["self"]},
        )

    def add_rule(
        self,
        endpoint: str,
        chain_id: str,
        target_name: str,
        target_type: str,
        rule: Optional[list[str]] = None,
        path: Optional[str] = None,
        chain_id_hex: Optional[bool] = None,
        wallet: Optional[str] = None,
        address: Optional[str] = None,
        timeout: Optional[str] = None,
    ) -> CommandResult:
        """Drop objects from the node's local storage

        Args:
        address: Address of wallet account
        chain-id: Assign ID to the parsed chain
        chain-id-hex: Flag to parse chain ID as hex
        endpoint: Remote node control address (as 'multiaddr' or '<host>:<port>')
        path: Path to encoded chain in JSON or binary format
        rule: Rule statement
        target-name: Resource name in APE resource name format
        target-type: Resource type(container/namespace)
        timeout: Timeout for an operation (default 15s)
        wallet: Path to the wallet or binary key

        Returns:
            Command`s result.
        """
        return self._execute(
            "control add-rule",
            **{param: value for param, value in locals().items() if param not in ["self"]},
        )

    def get_rule(
        self,
        endpoint: str,
        chain_id: str,
        target_name: str,
        target_type: str,
        chain_id_hex: Optional[bool] = None,
        wallet: Optional[str] = None,
        address: Optional[str] = None,
        timeout: Optional[str] = None,
    ) -> CommandResult:
        """Drop objects from the node's local storage

        Args:
        address string       Address of wallet account
        chain-id string      Chain id
        chain-id-hex         Flag to parse chain ID as hex
        endpoint string      Remote node control address (as 'multiaddr' or '<host>:<port>')
        target-name string   Resource name in APE resource name format
        target-type string   Resource type(container/namespace)
        timeout duration     Timeout for an operation (default 15s)
        wallet string        Path to the wallet or binary key

        Returns:
            Command`s result.
        """
        return self._execute(
            "control get-rule",
            **{param: value for param, value in locals().items() if param not in ["self"]},
        )

    def list_rules(
        self,
        endpoint: str,
        target_name: str,
        target_type: str,
        wallet: Optional[str] = None,
        address: Optional[str] = None,
        timeout: Optional[str] = None,
    ) -> CommandResult:
        """Drop objects from the node's local storage

        Args:
        address: Address of wallet account
        endpoint: Remote node control address (as 'multiaddr' or '<host>:<port>')
        target-name: Resource name in APE resource name format
        target-type: Resource type(container/namespace)
        timeout: Timeout for an operation (default 15s)
        wallet: Path to the wallet or binary key

        Returns:
            Command`s result.
        """
        return self._execute(
            "control list-rules",
            **{param: value for param, value in locals().items() if param not in ["self"]},
        )

    def list_targets(
        self,
        endpoint: str,
        chain_name: str,
        wallet: Optional[str] = None,
        address: Optional[str] = None,
        timeout: Optional[str] = None,
    ) -> CommandResult:
        """Drop objects from the node's local storage

        Args:
        address: Address of wallet account
        chain-name: Chain name(ingress|s3)
        endpoint: Remote node control address (as 'multiaddr' or '<host>:<port>')
        timeout: Timeout for an operation (default 15s)
        wallet: Path to the wallet or binary key

        Returns:
            Command`s result.
        """
        return self._execute(
            "control list-targets",
            **{param: value for param, value in locals().items() if param not in ["self"]},
        )

    def remove_rule(
        self,
        endpoint: str,
        chain_id: str,
        target_name: str,
        target_type: str,
        all: Optional[bool] = None,
        chain_id_hex: Optional[bool] = None,
        wallet: Optional[str] = None,
        address: Optional[str] = None,
        timeout: Optional[str] = None,
    ) -> CommandResult:
        """Drop objects from the node's local storage

        Args:
        address: Address of wallet account
        all: Remove all chains
        chain-id: Assign ID to the parsed chain
        chain-id-hex: Flag to parse chain ID as hex
        endpoint: Remote node control address (as 'multiaddr' or '<host>:<port>')
        target-name: Resource name in APE resource name format
        target-type: Resource type(container/namespace)
        timeout: Timeout for an operation (default 15s)
        wallet: Path to the wallet or binary key

        Returns:
            Command`s result.
        """
        return self._execute(
            "control remove-rule",
            **{param: value for param, value in locals().items() if param not in ["self"]},
        )
