from typing import Any

from frostfs_testlib import reporter
from frostfs_testlib.storage.cluster import ClusterNode
from frostfs_testlib.storage.controllers.cluster_state_controller import ClusterStateController, StateManager
from frostfs_testlib.storage.dataclasses.node_base import ServiceClass
from frostfs_testlib.testing import parallel


class ConfigStateManager(StateManager):
    def __init__(self, cluster_state_controller: ClusterStateController) -> None:
        super().__init__(cluster_state_controller)
        self.services_with_changed_config: set[tuple[ClusterNode, ServiceClass]] = set()
        self.cluster = self.csc.cluster

    @reporter.step("Change configuration for {service_type} on all nodes")
    def set_on_all_nodes(self, service_type: type[ServiceClass], values: dict[str, Any], sighup: bool = False):
        services = self.cluster.services(service_type)
        nodes = self.cluster.nodes(services)
        self.services_with_changed_config.update([(node, service_type) for node in nodes])

        if not sighup:
            self.csc.stop_services_of_type(service_type)

        parallel([node.config(service_type).set for node in nodes], values=values)
        if not sighup:
            self.csc.start_services_of_type(service_type)
        else:
            self.csc.sighup_services_of_type(service_type)

    @reporter.step("Change configuration for {service_type} on {node}")
    def set_on_node(self, node: ClusterNode, service_type: type[ServiceClass], values: dict[str, Any]):
        self.services_with_changed_config.add((node, service_type))

        self.csc.stop_service_of_type(node, service_type)
        node.config(service_type).set(values)
        self.csc.start_service_of_type(node, service_type)

    @reporter.step("Revert all configuration changes")
    def revert_all(self, sighup: bool = False):
        if not self.services_with_changed_config:
            return

        parallel(self._revert_svc, self.services_with_changed_config, sighup)
        self.services_with_changed_config.clear()

        if not sighup:
            self.csc.start_all_stopped_services()

    # TODO: parallel can't have multiple parallel_items :(
    @reporter.step("Revert all configuration {node_and_service}")
    def _revert_svc(self, node_and_service: tuple[ClusterNode, ServiceClass], sighup: bool = False):
        node, service_type = node_and_service
        service = node.service(service_type)

        if not sighup:
            self.csc.stop_service_of_type(node, service_type)

        node.config(service_type).revert()

        if sighup:
            service.send_signal_to_service("SIGHUP")
